@@ -11,37 +11,47 @@ import warnings
 from abc import ABC, abstractmethod
 from time import monotonic
 from types import TracebackType
-from typing import Any, AsyncIterator, Generic, cast
+from typing import Any, Generic, cast
 from weakref import ref
 from contextlib import asynccontextmanager
+from collections import deque
+from collections.abc import AsyncIterator
 
 from psycopg import AsyncConnection
 from psycopg import errors as e
 from psycopg.pq import TransactionStatus
 
-from .abc import ACT, AsyncConnectFailedCB, AsyncConnectionCB
+from .abc import ACT, AsyncConnectFailedCB, AsyncConnectionCB, AsyncConninfoParam
+from .abc import AsyncKwargsParam
 from .base import AttemptWithBackoff, BasePool
 from .errors import PoolClosed, PoolTimeout, TooManyRequests
-from ._compat import Deque, Self
+from ._compat import PSYCOPG_VERSION, AsyncPoolConnection, Self
 from ._acompat import ACondition, AEvent, ALock, AQueue, AWorker, agather, asleep
-from ._acompat import aspawn, current_task_name
+from ._acompat import aspawn, current_task_name, ensure_async
 from .sched_async import AsyncScheduler
 
 if True:  # ASYNC
     import asyncio
 
+    # The exceptions that we need to capture in order to keep the pool
+    # consistent and avoid losing connections on errors in callers code.
+    CLIENT_EXCEPTIONS = (Exception, asyncio.CancelledError)
+else:
+    CLIENT_EXCEPTIONS = Exception
+
+
 logger = logging.getLogger("psycopg.pool")
 
 
 class AsyncConnectionPool(Generic[ACT], BasePool):
-    _pool: Deque[ACT]
+    _pool: deque[ACT]
 
     def __init__(
         self,
-        conninfo: str = "",
+        conninfo: AsyncConninfoParam = "",
         *,
-        connection_class: type[ACT] = cast("type[ACT]", AsyncConnection),
-        kwargs: dict[str, Any] | None = None,
+        connection_class: type[ACT] = cast(type[ACT], AsyncConnection),
+        kwargs: AsyncKwargsParam | None = None,
         min_size: int = 4,
         max_size: int | None = None,
         open: bool | None = None,
@@ -49,6 +59,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
         check: AsyncConnectionCB[ACT] | None = None,
         reset: AsyncConnectionCB[ACT] | None = None,
         name: str | None = None,
+        close_returns: bool = False,
         timeout: float = 30.0,
         max_waiting: int = 0,
         max_lifetime: float = 60 * 60.0,
@@ -57,6 +68,18 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
         reconnect_failed: AsyncConnectFailedCB | None = None,
         num_workers: int = 3,
     ):
+        if close_returns and PSYCOPG_VERSION < (3, 3):
+            if connection_class is AsyncConnection:
+                connection_class = cast(type[ACT], AsyncPoolConnection)
+            else:
+                raise TypeError(
+                    "Using 'close_returns=True' and a non-standard 'connection_class'"
+                    " requires psycopg 3.3 or newer. Please check the docs at"
+                    " https://www.psycopg.org/psycopg3/docs/advanced/pool.html"
+                    "#pool-sqlalchemy for a workaround."
+                )
+        self.conninfo = conninfo
+        self.kwargs = kwargs
         self.connection_class = connection_class
         self._check = check
         self._configure = configure
@@ -70,7 +93,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
         self._sched: AsyncScheduler
         self._tasks: AQueue[MaintenanceTask]
 
-        self._waiting = Deque[WaitingClient[ACT]]()
+        self._waiting = deque[WaitingClient[ACT]]()
 
         # to notify that the pool is full
         self._pool_full_event: AEvent | None = None
@@ -79,11 +102,10 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
         self._workers: list[AWorker] = []
 
         super().__init__(
-            conninfo,
-            kwargs=kwargs,
             min_size=min_size,
             max_size=max_size,
             name=name,
+            close_returns=close_returns,
             timeout=timeout,
             max_waiting=max_waiting,
             max_lifetime=max_lifetime,
@@ -237,7 +259,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
             conn = await self._getconn_unchecked(deadline - monotonic())
             try:
                 await self._check_connection(conn)
-            except Exception:
+            except CLIENT_EXCEPTIONS:
                 await self._putconn(conn, from_getconn=True)
             else:
                 logger.info("connection given by %r", self.name)
@@ -275,7 +297,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
         if not conn:
             try:
                 conn = await pos.wait(timeout=timeout)
-            except Exception:
+            except CLIENT_EXCEPTIONS:
                 self._stats[self._REQUESTS_ERRORS] += 1
                 raise
             finally:
@@ -312,7 +334,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
             return
         try:
             await self._check(conn)
-        except Exception as e:
+        except CLIENT_EXCEPTIONS as e:
             logger.info("connection failed check: %s", e)
             raise
 
@@ -341,6 +363,26 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
 
         await self._putconn(conn, from_getconn=False)
 
+    async def drain(self) -> None:
+        """
+        Remove all the connections from the pool and create new ones.
+
+        If a connection is currently out of the pool it will be closed when
+        returned to the pool and replaced with a new one.
+
+        This method is useful to force a connection re-configuration, for
+        example when the adapters map changes after the pool was created.
+        """
+        async with self._lock:
+            conns = list(self._pool)
+            self._pool.clear()
+            self._drained_at = monotonic()
+
+        # Close the connection already in the pool, open new ones.
+        for conn in conns:
+            await self._close_connection(conn)
+            self.run_task(AddConnection(self))
+
     async def _putconn(self, conn: ACT, from_getconn: bool) -> None:
         # Use a worker to perform eventual maintenance work in a separate task
         if self._reset:
@@ -358,8 +400,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
         if not self._closed:
             return False
 
-        conn._pool = None
-        await conn.close()
+        await self._close_connection(conn)
         return True
 
     async def open(self, wait: bool = False, timeout: float = 30.0) -> None:
@@ -473,8 +514,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
 
         # Close the connections that were still in the pool
         for conn in connections:
-            conn._pool = None
-            await conn.close()
+            await self._close_connection(conn)
 
         # Signal to eventual clients in the queue that business is closed.
         for pos in waiting:
@@ -546,15 +586,14 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
             # Check for expired connections
             if conn._expire_at <= monotonic():
                 logger.info("discarding expired connection %s", conn)
-                conn._pool = None
-                await conn.close()
+                await self._close_connection(conn)
                 self.run_task(AddConnection(self))
                 continue
 
             # Check for broken connections
             try:
                 await self.check_connection(conn)
-            except Exception:
+            except CLIENT_EXCEPTIONS:
                 self._stats[self._CONNECTIONS_LOST] += 1
                 logger.warning("discarding broken connection: %s", conn)
                 self.run_task(AddConnection(self))
@@ -598,15 +637,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
         if not self._reconnect_failed:
             return
 
-        if True:  # ASYNC
-            import inspect
-
-            if inspect.iscoroutinefunction(self._reconnect_failed):
-                await self._reconnect_failed(self)
-            else:
-                self._reconnect_failed(self)
-        else:
-            self._reconnect_failed(self)
+        await ensure_async(self._reconnect_failed, self)
 
     def run_task(self, task: MaintenanceTask) -> None:
         """Run a maintenance task in a worker."""
@@ -633,7 +664,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
             # Run the task. Make sure don't die in the attempt.
             try:
                 await task.run()
-            except Exception as ex:
+            except CLIENT_EXCEPTIONS as ex:
                 logger.warning(
                     "task run %s failed: %s: %s", task, ex.__class__.__name__, ex
                 )
@@ -641,14 +672,15 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
     async def _connect(self, timeout: float | None = None) -> ACT:
         """Return a new connection configured for the pool."""
         self._stats[self._CONNECTIONS_NUM] += 1
-        kwargs = self.kwargs
+        conninfo = await self._resolve_conninfo()
+        kwargs = await self._resolve_kwargs()
         if timeout:
             kwargs = kwargs.copy()
             kwargs["connect_timeout"] = max(round(timeout), 1)
         t0 = monotonic()
         try:
-            conn = await self.connection_class.connect(self.conninfo, **kwargs)
-        except Exception:
+            conn = await self.connection_class.connect(conninfo, **kwargs)
+        except CLIENT_EXCEPTIONS:
             self._stats[self._CONNECTIONS_ERRORS] += 1
             raise
         else:
@@ -670,6 +702,23 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
         self._set_connection_expiry_date(conn)
         return conn
 
+    async def _resolve_conninfo(self) -> str:
+        """Resolve conninfo (static string, sync callable, or async callable)."""
+        if callable(self.conninfo):
+            return await ensure_async(self.conninfo)
+
+        return self.conninfo or ""
+
+    async def _resolve_kwargs(self) -> dict[str, Any]:
+        """Resolve kwargs (static dict, sync callable, or async callable)."""
+        if not self.kwargs:
+            return {}
+
+        if callable(self.kwargs):
+            return await ensure_async(self.kwargs)
+
+        return self.kwargs
+
     async def _add_connection(
         self, attempt: AttemptWithBackoff | None, growing: bool = False
     ) -> None:
@@ -686,8 +735,8 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
 
         try:
             conn = await self._connect()
-        except Exception as ex:
-            logger.warning(f"error connecting in {self.name!r}: {ex}")
+        except CLIENT_EXCEPTIONS as ex:
+            logger.warning("error connecting in %r: %s", self.name, ex)
             if attempt.time_to_give_up(now):
                 logger.warning(
                     "reconnection attempt in pool %r failed after %s sec",
@@ -744,10 +793,9 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
                 return
 
         # Check if the connection is past its best before date
-        if conn._expire_at <= monotonic():
+        if conn._created_at <= self._drained_at or conn._expire_at <= monotonic():
             logger.info("discarding expired connection")
-            conn._pool = None
-            await conn.close()
+            await self._close_connection(conn)
             self.run_task(AddConnection(self))
             return
 
@@ -772,7 +820,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
         # between here and entering the lock. Therefore we will make another
         # check later.
         if self._closed:
-            await conn.close()
+            await self._close_connection(conn)
             return
 
         # Critical section: if there is a client waiting give it the connection
@@ -784,7 +832,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
             # this connection while the main process is closing the pool.
             # Now that we are in the critical section we know for real.
             if self._closed:
-                await conn.close()
+                await self._close_connection(conn)
                 return
 
             while self._waiting:
@@ -816,21 +864,19 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
             logger.warning("rolling back returned connection: %s", conn)
             try:
                 await conn.rollback()
-            except Exception as ex:
+            except CLIENT_EXCEPTIONS as ex:
                 logger.warning(
                     "rollback failed: %s: %s. Discarding connection %s",
                     ex.__class__.__name__,
                     ex,
                     conn,
                 )
-                conn._pool = None
-                await conn.close()
+                await self._close_connection(conn)
 
         elif status == TransactionStatus.ACTIVE:
             # Connection returned during an operation. Bad... just close it.
             logger.warning("closing returned connection: %s", conn)
-            conn._pool = None
-            await conn.close()
+            await self._close_connection(conn)
 
         if self._reset:
             try:
@@ -841,10 +887,13 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
                         f"connection left in status {sname} by reset function"
                         f" {self._reset}: discarded"
                     )
-            except Exception as ex:
-                logger.warning(f"error resetting connection: {ex}")
-                conn._pool = None
-                await conn.close()
+            except CLIENT_EXCEPTIONS as ex:
+                logger.warning("error resetting connection: %s", ex)
+                await self._close_connection(conn)
+
+    async def _close_connection(self, conn: ACT) -> None:
+        conn._pool = None
+        await conn.close()
 
     async def _shrink_pool(self) -> None:
         to_close: ACT | None = None
@@ -869,8 +918,7 @@ class AsyncConnectionPool(Generic[ACT], BasePool):
                 nconns_min,
                 self.max_idle,
             )
-            to_close._pool = None
-            await to_close.close()
+            await self._close_connection(to_close)
 
     def _get_measures(self) -> dict[str, int]:
         rv = super()._get_measures()
@@ -906,7 +954,7 @@ class WaitingClient(Generic[ACT]):
                         self.error = PoolTimeout(
                             f"couldn't get a connection after {timeout:.2f} sec"
                         )
-                except BaseException as ex:
+                except CLIENT_EXCEPTIONS as ex:
                     self.error = ex
 
         if self.conn:
